import requests
import pandas as pd
from ._constants import CATEGORIES, TOPIC_TAGS


class GetQuestionsList:
    """A class to scrape the list of questions, their topic tags, and company tags.

    Args:
        limit (int, optional): The maximum number of questions to query for from Leetcode's graphql API. Defaults to 10,000.
    """

    def __init__(self, limit: int = 10_000):
        self.limit = limit

    def scrape(self):
        """Scrapes LeetCode data including company tags, questions, question topics,
        and categories.
        """
        self._scrape_companies()
        self._scrape_questions_list()
        self._extract_question_topics()
        self._get_categories_and_topicTags_lists()
        self._scrape_question_category()
        self._add_category_to_questions_list()

    def _get_categories_and_topicTags_lists(self):
        """Get the categories and topic tags of LeetCode problems and store them in the
        'categories' and 'topicTags' attribute respectively."""
        print("Getting Categories ... ", end="")
        # List of problem categories
        self.categories = pd.DataFrame.from_records(CATEGORIES)
        print("Done")
        # List of problem topic tags
        print("Scraping Topic Tags ... ", end="")
        self.topicTags = pd.DataFrame.from_records(TOPIC_TAGS)
        print("Done")

    def _scrape_question_category(self):
        """Scrape the category of each question and store it in the 'questionCategory' dataframe."""
        print("Extracting question category ... ", end="")
        categories_data = []
        for category in self.categories["slug"].values:
            data = {
                "query": """query problemsetQuestionList($categorySlug: String, $limit: Int, $skip: Int, $filters: QuestionListFilterInput) {
                        problemsetQuestionList: questionList(
                            categorySlug: $categorySlug
                            limit: $limit
                            skip: $skip
                            filters: $filters
                        ) {
                            questions: data {
                                QID: questionFrontendId
                            }
                        }
                    }
                """,
                "variables": {
                    "categorySlug": category,
                    "skip": 0,
                    "limit": self.limit,
                    "filters": {},
                },
            }

            r = requests.post("https://leetcode.com/graphql", json=data).json()
            categories = pd.json_normalize(
                r["data"]["problemsetQuestionList"]["questions"]
            )
            categories["categorySlug"] = category
            categories_data.append(categories)
        self.questionCategory = pd.concat(categories_data, axis=0, ignore_index=True)
        print("Done")

    def _scrape_questions_list(self):
        """
        Scrapes the list of questions from leetcode.com and store them in the 'questions'
        dataframe. The columns include the question QID, acceptance rate, difficulty,
        title, titleSlug, and topic tags. It also has a column indicating
        whether the question is available only to Leetcode's paying customers.
        """
        print("Scraping questions list ... ", end="")
        data = {
            "query": """query problemsetQuestionList($categorySlug: String, $limit: Int, $skip: Int, $filters: QuestionListFilterInput) {
                    problemsetQuestionList: questionList(
                        categorySlug: $categorySlug
                        limit: $limit
                        skip: $skip
                        filters: $filters
                    ) {
                        total: totalNum
                        questions: data {
                            acceptanceRate: acRate
                            difficulty
                            QID: questionFrontendId
                            paidOnly: isPaidOnly
                            title
                            titleSlug
                            topicTags {
                                slug
                            }
                        }
                    }
                }
            """,
            "variables": {
                "categorySlug": "",
                "skip": 0,
                "limit": self.limit,
                "filters": {},
            },
        }

        r = requests.post("https://leetcode.com/graphql", json=data).json()
        self.questions = pd.json_normalize(
            r["data"]["problemsetQuestionList"]["questions"]
        )[
            [
                "QID",
                "title",
                "titleSlug",
                "difficulty",
                "acceptanceRate",
                "paidOnly",
                "topicTags",
            ]
        ]
        self.questions["topicTags"] = self.questions["topicTags"].apply(
            lambda w: [tag["slug"] for tag in w]
        )
        print("Done")

    def _extract_question_topics(self):
        """Create a table with the edge list of questions and topic tags."""
        print("Extracting question topics ... ", end="")
        self.questionTopics = (
            self.questions[["QID", "topicTags"]]
            .rename(columns={"topicTags": "tagSlug"})
            .explode("tagSlug", ignore_index=True)
        ).dropna()
        print("Done")

    def _scrape_companies(self):
        """Scrape the company tags of each question. This always returns an empty
        dataframe as this is a paid only feature."""
        print("Scraping companies ... ", end="")
        data = {
            "query": """query questionCompanyTags {
                    companyTags {
                        name
                        slug
                        questionCount
                    }
                }
            """,
            "variables": {},
        }
        r = requests.post("https://leetcode.com/graphql", json=data).json()
        self.companies = pd.json_normalize(r["data"]["companyTags"])
        print("Done")

    def _add_category_to_questions_list(self):
        """Adds the `topicTags` column containing the comma-separated string of
        the list of topic tags relevant to the given questions and the `category`
        column that includes the category relevant to the given question"""
        self.questions["topicTags"] = self.questions["topicTags"].apply(
            lambda w: ",".join(w)
        )
        self.questions = self.questions.join(
            self.questionCategory.set_index("QID"), on="QID"
        )

    def to_csv(self, directory_path: str) -> None:
        """A method to export the scraped data into csv files in preparation for
        injection into a database.

        Args:
            directory_path (str): The directory path to export the scraped data into.
        """
        self.companies.to_csv(directory_path + "companies.csv", index=False)
        self.questions["QID"] = self.questions["QID"].astype(int)
        self.questions.to_csv(directory_path + "questions.csv", index=False)
        self.questionTopics.to_csv(
            directory_path + "questionTopics.csv", index=True, index_label="id"
        )
        self.categories.to_csv(directory_path + "categories.csv", index=False)
        self.topicTags.to_csv(directory_path + "topicTags.csv", index=False)
        self.questionCategory.to_csv(
            directory_path + "questionCategory.csv", index=True, index_label="id"
        )



import ast
import pickle
import re

import numpy as np
import pandas as pd
import sqlalchemy
from sqlalchemy import MetaData, update
from tqdm import tqdm

from .GetQuestionInfo import GetQuestionInfo

"""A set of helper functions to transform and query the scraped data."""


def combine_list_and_info(
    list_df: pd.DataFrame, info_df: pd.DataFrame, save_to: str = ""
) -> pd.DataFrame:
    """
    Combines the questions list dataframe with the questions info dataframe by QID
    and return the combined dataframe where each record contains the QID, title,
    titleSlug, hints, difficulty, acceptance rate, similar questions, topic tags,
    category, code stubs, body, and companies.

    Args:
        list_df (pd.DataFrame): Dataframe that contains the list of questions.
        info_df (pd.DataFrame): Dataframe that contains the additional information of questions.
        save_to (str): If provided, saves the resulting dataframe to a json file with the given name. Defaults to "".

    Returns:
        pd.DataFrame: The combined dataframe.
    """
    questions = pd.concat(
        [list_df.set_index("QID"), info_df.set_index("QID")], axis=1
    ).reset_index()
    questions["Hints"] = questions["Hints"].fillna("").apply(list)
    questions["SimilarQuestions"] = questions["SimilarQuestions"].fillna("").apply(list)
    questions["SimilarQuestions"] = questions["SimilarQuestions"].apply(
        lambda w: [int(q) for q in w]
    )
    questions["Code"] = questions["Code"].fillna("")
    questions["Body"] = questions["Body"].fillna("")
    questions["Companies"] = (
        questions["Companies"].fillna(np.nan).replace([np.nan], [None])
    )
    if len(save_to) > 0:
        questions.to_json(save_to, orient="records")
    return questions


def get_all_questions_body(
    titleSlugs, isPaidOnlyList, save_to: str = "../example/data/dump.pickle"
) -> list[dict[str, str | int]]:
    """
    Get the body of all questions in the list and save it to a file.

    Args:
        titleSlugs (List[str]): A list of titleSlugs of questions.
        isPaidOnlyList (List[bool]): A list of isPaidOnly property of questions, indicating whether the question is subscriber only.
        save_to (str):  The path to save the scraped data.

    Returns:
        list[dict[str, str|int]]: A list of dictionaries containing the question information
    """
    questions_info_list = []
    for i, (titleSlug, paidOnly) in enumerate(tqdm(zip(titleSlugs, isPaidOnlyList))):
        if not paidOnly:
            questions_info_list.append(GetQuestionInfo(titleSlug).scrape())
        if i % 10 == 0:
            with open(save_to, "wb") as f:
                pickle.dump(questions_info_list, f)

    with open(save_to, "wb") as f:
        pickle.dump(questions_info_list, f)
    return questions_info_list


def extract_solutions(filename: str) -> dict:
    """
    Extract solutions from a given python file.

    Args:
        filename (str): The path of the file to extract solutions from. It should be of the following form `q_{{LEETCODE_QID}}_{{LEETCODE_TITLE}}.py`. Furthermore, the python script file should have the solution method in a class named `Solution`.

    Raises:
        ValueError: When the filename does not follow the required convention of `q_{{LEETCODE_QID}}_ {{LEETCODE_TITLE}}.py`.
        ValueError: When the provided python file does not have a class named Solution.

    Returns:
        dict: A dictionary containing the question id and a list of solutions, each solution contains an id, code [and docs].
    """
    qid = re.search("q_(.*)_", filename)
    if qid is not None:
        qid = int(qid.group(1))
    else:
        raise ValueError(
            f"""Invalid filename. Filename should be of the form `q_{{LEETCODE_QID}}_{{LEETCODE_TITLE}}.py`. Received `{filename}` instead."""
        )
    with open(filename) as fd:
        file_contents = fd.read()
    module = ast.parse(file_contents)
    class_definition = [
        node
        for node in module.body
        if isinstance(node, ast.ClassDef) and node.name == "Solution"
    ]
    if not class_definition:
        raise ValueError("The provided python file should have a class named Solution.")
    method_definitions = [
        node for node in class_definition[0].body if isinstance(node, ast.FunctionDef)
    ]

    solutions = [
        {
            "id": idx,
            "code": ast.get_source_segment(file_contents, f, padded=True),
            "docs": ast.get_docstring(f, clean=True),
        }
        for idx, f in enumerate(method_definitions)
    ]

    return {qid: solutions}


def upload_solutions(
    engine: sqlalchemy.engine.Engine,
    row_id: int,
    solutions: dict,
    table_name: str = "questions",
    col_name: str = "solutions",
) -> None:
    """Upload solutions to the corresponding row of a table in a database.

    Args:
        engine (sqlalchemy.engine.Engine): The sqlalchemy engine used to connect to the database.
        row_id (int): The id of the row that the solutions should be uploaded to.
        solutions (dict): The solutions to be uploaded.
        table_name (str): The name of the table to upload the solutions to. Defaults to "questions".
        col_name (str): The name of the column to upload the solutions to. Defaults to "solutions".
    """
    varss = MetaData(bind=engine)
    MetaData.reflect(varss)
    questions = varss.tables[table_name]
    engine.execute(
        update(questions).where(questions.c.QID == row_id).values({col_name: solutions})
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['leetscrape']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.46,<2.0.0',
 'numpy>=1.24.1,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'psycopg2>=2.9.5,<3.0.0',
 'requests>=2.28.2,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'leetscrape',
    'version': '0.1.3',
    'description': 'Introducing LeetScrape - a powerful and efficient Python package designed to scrape problem statements and their topic and company tags, difficulty, test cases, hints, and code stubs from LeetCode.com. Easily download and save LeetCode problems to your local machine, making it convenient for offline practice and studying. It is perfect for anyone preparing for coding interviews. With the LeetScrape, you can boost your coding skills and improve your chances of landing your dream job.',
    'long_description': '# Leetcode Questions Scraper\n\n[![Python application](https://github.com/nikhil-ravi/LeetcodeScraper/actions/workflows/python-app.yml/badge.svg)](https://github.com/nikhil-ravi/LeetcodeScraper/actions/workflows/python-app.yml) [![deploy-docs](https://github.com/nikhil-ravi/LeetScrape/actions/workflows/deploy-docs.yml/badge.svg)](https://leetscrape.chowkabhara.com) [![PYPI](https://img.shields.io/pypi/v/leetscrape)](https://pypi.org/project/leetscrape/)\n\nIntroducing the LeetScrape - a powerful and efficient Python package designed to scrape problem statements and basic test cases from LeetCode.com. With this package, you can easily download and save LeetCode problems to your local machine, making it convenient for offline practice and studying. It is perfect for software engineers and students preparing for coding interviews. The package is lightweight, easy to use and can be integrated with other tools and IDEs. With the LeetScrape, you can boost your coding skills and improve your chances of landing your dream job.\n\nUse this package to get the list of Leetcode questions, their topic and company tags, difficulty, question body (including test cases, constraints, hints), and code stubs in any of the available programming languages.\n\n## Usage\n\nImport the relevant classes from the [`leetcode`](/src/leetcode/) package:\n\n```python\nfrom leetscrape.GetQuestionsList import GetQuestionsList\nfrom leetscrape.GetQuestionInfo import GetQuestionInfo\nfrom leetscrape.utils import combine_list_and_info, get_all_questions_body\n```\n\nGet the list of questions, companies, topic tags, categories using the [`GetQuestionsList`](/src/leetcode/GetQuestionsList.py) class:\n\n```python\nls = GetQuestionsList()\nls.scrape() # Scrape the list of questions\nls.to_csv(directory_path="../data/") # Save the scraped tables to a directory\n```\n\n> **Warning**\n> The default ALL_JSON_URL in the [`GetQuestionsList`](/src/leetcode/GetQuestionsList.py) class might be out-of-date. Please update it by going to https://leetcode.com/problemset/all/ and exploring the Networks tab for a query returning all.json.\n\nQuery individual question\'s information such as the body, test cases, constraints, hints, code stubs, and company tags using the [`GetQuestionInfo`](/src/leetcode/GetQuestionInfo.py) class:\n\n```python\n# This table can be generated using the previous commnd\nquestions_info = pd.read_csv("../data/questions.csv")\n\n# Scrape question body\nquestions_body_list = get_all_questions_body(\n    questions_info["titleSlug"].tolist(),\n    questions_info["paidOnly"].tolist(),\n    save_to="../data/questionBody.pickle",\n)\n\n# Save to a pandas dataframe\nquestions_body = pd.DataFrame(\n    questions_body_list\n).drop(columns=["titleSlug"])\nquestions_body["QID"] = questions_body["QID"].astype(int)\n```\n\n> **Note**\n> The above code stub is time consuming (10+ minutes) since there are 2500+ questions.\n\nCreate a new dataframe with all the questions and their metadata and body information.\n\n```python\nquestions = combine_list_and_info(\n    info_df = questions_body, list_df=ls.questions, save_to="../data/all.json"\n)\n```\n\nCreate a PostgreSQL database using the [SQL](/example/sql/create.sql) dump and insert data using `sqlalchemy`.\n\n```python\nfrom sqlalchemy import create_engine\nfrom sqlalchemy.orm import sessionmaker\n\nengine = create_engine("<database_connection_string>", echo=True)\nquestions.to_sql(con=engine, name="questions", if_exists="append", index=False)\n# Repeat the same for tables ls.topicTags, ls.categories,\n# ls.companies, # ls.questionTopics, and ls.questionCategory\n```\n\nUse the `queried_questions_list` PostgreSQL function (defined in the SQL dump) to query for questions containy query terms:\n\n```sql\nselect * from queried_questions_list(\'<query term>\');\n```\n\nUse the `all_questions_list` PostgreSQL function (defined in the SQL dump) to query for all the questions in the database:\n\n```sql\nselect * from all_questions_list();\n```\n\nUse the `get_similar_questions` PostgreSQL function (defined in the SQL dump) to query for all questions similar to a given question:\n\n```sql\nselect * from get_similar_questions(<QuestionID>);\n```\n\nUse the [`extract_solutions`](/src/leetcode/utils.py:) method to extract solution code stubs from your python script. Note that the solution method should be a part of a class named `Solution` (see [here](/example/solutions/q_0001_TwoSum.py) for an example):\n\n```python\n# Returns a dict of the form {QuestionID: solutions}\nsolutions = extract_solutions(filename=<path_to_python_script>)\n```\n\nUse the [`upload_solutions`](/src/leetcode/utils.py:) method to upload the extracted solution code stubs from your python script to the PosgreSQL database.\n\n```python\nupload_solutions(engine=<sqlalchemy_engine>, row_id = <row_id_in_table>, solutions: <solutions_dict>)\n```\n',
    'author': 'Nikhil Ravi',
    'author_email': 'nr337@cornell.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nikhil-ravi/LeetScrape',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import requests
import pandas as pd
import json
import time

# Leetcode's graphql api endpoint
BASE_URL = "https://leetcode.com/graphql"


class GetQuestionInfo:
    """
    A class to acquire the statement, constraints, hints, basic test cases, related questions, and code stubs of the given question.

    Args:
        titleSlug (str): The title slug of the question.
        questions_info_path (str): The filename of the file containing the list of all questions containing their QID and titleSlug.
    """

    def __init__(
        self, titleSlug: str, questions_info_path: str = "../example/data/questions.csv"
    ):
        self.titleSlug = titleSlug
        self.questions_info = pd.read_csv(
            questions_info_path, usecols=["QID", "titleSlug"], index_col="titleSlug"
        )

    def scrape(self) -> dict:
        """This method calls the Leetcode graphql api to query for the hints, companyTags (currently returning null as this is a premium feature), code snippets, and content of the question.

        Raises:
            ValueError: When the connection to Leetcode's graphql api is not established.

        Returns:
            dict: Dictionary containing the QID, titleSlug, Hints, Companies, Similar Questions, Code stubs, and the body of the question.
        """
        data = {
            "query": """query questionHints($titleSlug: String!) {
                question(titleSlug: $titleSlug) {
                    hints
                    companyTags {
                        name
                        slug
                        imgUrl
                    }
                    similarQuestions
                    codeSnippets {
                        lang
                        langSlug
                        code
                    }
                    content
                }
            }
        """,
            "variables": {"titleSlug": self.titleSlug},
        }
        req = requests.post(BASE_URL, json=data)
        if req.status_code == 404:
            raise ValueError(self.titleSlug)
        while req.status_code == 429:
            time.sleep(10)
            req = requests.post(BASE_URL, json=data)
            if req.status_code == 404:
                raise ValueError(self.titleSlug)
        self.req = req.json()
        return {
            "QID": self.questions_info.loc[self.titleSlug].QID,
            "titleSlug": self.titleSlug,
            "Hints": self.req["data"]["question"]["hints"],
            "Companies": self.req["data"]["question"]["companyTags"],
            "SimilarQuestions": self.get_similar_questions(),
            "Code": self.get_code_snippet(),
            "Body": self.req["data"]["question"]["content"],
        }

    # Similar questions
    def get_similar_questions(self) -> list[int]:
        """A helper method to extract the list of similar questions of the
        given question.

        Returns:
            list[int]: The list of QIDs of the questions similar to the given question.
        """
        similar_questions = []
        for qs in json.loads(self.req["data"]["question"]["similarQuestions"]):
            similar_questions.append(self.questions_info.loc[qs["titleSlug"]].QID)
        return similar_questions

    # Code Snippet
    def get_code_snippet(self) -> str:
        """A helper method to extract the code snippets from the query response.
        Currently, this method returns the Python3 code snippet if available,
        else it returns a barebones Python3 code snippet with the class name and
        method named after the titleSlug.

        Returns:
            str: Python3 code snippet
        """
        python_code_snippet = [
            code_snippet
            for code_snippet in self.req["data"]["question"]["codeSnippets"]
            if code_snippet["langSlug"] == "python3"
        ]
        if len(python_code_snippet) > 0:
            return python_code_snippet[0]["code"]
        else:
            return f"# This question has no Python code stub\nclass Solution:\n    def {self.titleSlug}(self) -> Any:\n      "



# List of problem categories
CATEGORIES = [
    {"slug": "algorithms", "name": "Algorithms"},
    {"slug": "database", "name": "Database"},
    {"slug": "shell", "name": "Shell"},
    {"slug": "concurrency", "name": "Concurrency"},
]

# List of topic tags
TOPIC_TAGS = [
    {"slug": "array", "name": "Array"},
    {"slug": "string", "name": "String"},
    {"slug": "hash-table", "name": "Hash Table"},
    {"slug": "dynamic-programming", "name": "Dynamic Programming"},
    {"slug": "math", "name": "Math"},
    {"slug": "sorting", "name": "Sorting"},
    {"slug": "greedy", "name": "Greedy"},
    {"slug": "depth-first-search", "name": "Depth-First Search"},
    {"slug": "database", "name": "Database"},
    {"slug": "breadth-first-search", "name": "Breadth-First Search"},
    {"slug": "binary-search", "name": "Binary Search"},
    {"slug": "tree", "name": "Tree"},
    {"slug": "matrix", "name": "Matrix"},
    {"slug": "binary-tree", "name": "Binary Tree"},
    {"slug": "two-pointers", "name": "Two Pointers"},
    {"slug": "bit-manipulation", "name": "Bit Manipulation"},
    {"slug": "stack", "name": "Stack"},
    {"slug": "heap-priority-queue", "name": "Heap (Priority Queue)"},
    {"slug": "design", "name": "Design"},
    {"slug": "graph", "name": "Graph"},
    {"slug": "prefix-sum", "name": "Prefix Sum"},
    {"slug": "simulation", "name": "Simulation"},
    {"slug": "counting", "name": "Counting"},
    {"slug": "backtracking", "name": "Backtracking"},
    {"slug": "sliding-window", "name": "Sliding Window"},
    {"slug": "union-find", "name": "Union Find"},
    {"slug": "linked-list", "name": "Linked List"},
    {"slug": "ordered-set", "name": "Ordered Set"},
    {"slug": "monotonic-stack", "name": "Monotonic Stack"},
    {"slug": "enumeration", "name": "Enumeration"},
    {"slug": "recursion", "name": "Recursion"},
    {"slug": "trie", "name": "Trie"},
    {"slug": "divide-and-conquer", "name": "Divide and Conquer"},
    {"slug": "binary-search-tree", "name": "Binary Search Tree"},
    {"slug": "bitmask", "name": "Bitmask"},
    {"slug": "queue", "name": "Queue"},
    {"slug": "memoization", "name": "Memoization"},
    {"slug": "geometry", "name": "Geometry"},
    {"slug": "segment-tree", "name": "Segment Tree"},
    {"slug": "topological-sort", "name": "Topological Sort"},
    {"slug": "number-theory", "name": "Number Theory"},
    {"slug": "hash-function", "name": "Hash Function"},
    {"slug": "binary-indexed-tree", "name": "Binary Indexed Tree"},
    {"slug": "game-theory", "name": "Game Theory"},
    {"slug": "data-stream", "name": "Data Stream"},
    {"slug": "interactive", "name": "Interactive"},
    {"slug": "string-matching", "name": "String Matching"},
    {"slug": "rolling-hash", "name": "Rolling Hash"},
    {"slug": "shortest-path", "name": "Shortest Path"},
    {"slug": "combinatorics", "name": "Combinatorics"},
    {"slug": "randomized", "name": "Randomized"},
    {"slug": "brainteaser", "name": "Brainteaser"},
    {"slug": "monotonic-queue", "name": "Monotonic Queue"},
    {"slug": "merge-sort", "name": "Merge Sort"},
    {"slug": "iterator", "name": "Iterator"},
    {"slug": "concurrency", "name": "Concurrency"},
    {"slug": "doubly-linked-list", "name": "Doubly-Linked List"},
    {"slug": "probability-and-statistics", "name": "Probability and Statistics"},
    {"slug": "quickselect", "name": "Quickselect"},
    {"slug": "bucket-sort", "name": "Bucket Sort"},
    {"slug": "suffix-array", "name": "Suffix Array"},
    {"slug": "minimum-spanning-tree", "name": "Minimum Spanning Tree"},
    {"slug": "counting-sort", "name": "Counting Sort"},
    {"slug": "shell", "name": "Shell"},
    {"slug": "line-sweep", "name": "Line Sweep"},
    {"slug": "reservoir-sampling", "name": "Reservoir Sampling"},
    {"slug": "eulerian-circuit", "name": "Eulerian Circuit"},
    {"slug": "radix-sort", "name": "Radix Sort"},
    {"slug": "strongly-connected-component", "name": "Strongly Connected Component"},
    {"slug": "rejection-sampling", "name": "Rejection Sampling"},
    {"slug": "biconnected-component", "name": "Biconnected Component"},
]
